// setup global configuration
if (!ul) var ul = {}; // global configuration
if (!ul.hl) ul.hl = {}; // code highlight configuration
if (!ul.hl.hl_def) ul.hl.hl_def = '1'; // default code highlight state
if (!ul.hl.select_id) ul.hl.select_id = 'id_select_hl';  // code highlight select element Id
if (!ul.params) ul.params = {}; // parameters
if (!ul.params.hl) ul.params.hl = ul.hl.hl_def; // code highlight state (0 - don't highlight, 1 - highlight)
if ((ul.params.hl.toLowerCase() == 'off') || (ul.params.hl.toLowerCase() == 'no') || (ul.params.hl.toLowerCase() == 'n')) ul.params.hl = 0;
if ((ul.params.hl.toLowerCase() == 'on') || (ul.params.hl.toLowerCase() == 'yes') || (ul.params.hl.toLowerCase() == 'y')) ul.params.hl = 1;
if ((ul.params.hl != 0) && (ul.params.hl != 1)) ul.params.hl = ul.hl.hl_def; // filter wrong highlighting values
if (!ul.hl.re) ul.hl.re = {}; // common regexps
if (!ul.hl.attr_code) ul.hl.attr_code = 'code'; // source code attribute added to source code containing element
if (!ul.hl.attr_raw_text) ul.hl.attr_raw_text = 'raw_text'; // attribute containing unhighlighted raw text

// common Regexps
ul.hl.re = {
  'comm_c'          : '\\/\\*.*\\*\\/', // C comment
  'comm_cpp'        : '\\/\\/.*$',      // C++ comment
  'str' : '[\'\"](\\\\?.)*?[\'\"]', // single/double quoted string with escaped symbols
  //'str_single'      : '\'[^\'\r\n]*\'', // single-line single-quoted string
  'str_trsingle'    : '\'\'\'[\\w\\W\\r\\n]*?\'\'\'', // multiline triple-single-quoted string
  'str_trdouble'    : '\"\"\"[\\w\\W\\r\\n]*?\"\"\"', // multiline triple-double-quoted string
  'alpha'           : 'a-zA-Z',  // alpha characters
  'num'             : '\\d',     // numeric characters
  'underscore'      : '_',       // underscore
  'alphaus'         : ul.hl.re.alpha + ul.hl.re.underscore, // alpha + underscore
  'alphanum'        : ul.hl.re.alpha + ul.hl.re.num,        // alpha + digits
  'alphanumus'      : ul.hl.re.alphaus + ul.hl.re.num,      // alphanumeric + underscore
  'identifier'      : '[' + ul.hl.re.alphaus + ']' + '[' + ul.hl.re.alphanumus + ']' + '*',
  'brackets_round'  : '[\\(\\)]', // round brackets
  'brackets_square' : '[\\[\\]]', // square brackets
  'braces'          : '[\\{\\}]', // braces
  'brackets_all'    : '[\\(\\)\\[\\]\\{\\}]', // all brackets
  'num_float'       : '[-\\+]?([0-9]+\\.?[0-9]*|\\.[0-9]+)([eE][-\\+]?[0-9]+)?', // floating-point number
  'num_int'         : '(?!\\W)(0[xX])?(0[bB])?[\\da-fA-F]+[lL]?' // integer number
};

// highlight schemes
// cn - language element style class name, rs - starting regexp, re - ending regexp, words - array of words
ul.hl.schemes =
 {'python' : { // Python code scheme
     'comment':   {'cn': 'hl_comment',  'rs': '#.*$'}
    ,'string':    {'cn': 'hl_string',   'rs': [ul.hl.re.str_trsingle, ul.hl.re.str_trdouble,'[uU]?' + ul.hl.re.str].join('|')}
    ,'bracket':   {'cn': 'hl_bracket',  'rs': ul.hl.re.brackets_all}
    ,'floating':  {'cn': 'hl_number',   'rs': ul.hl.re.num_float}
    ,'special':   {'cn': 'hl_special',  'rs': '\\.?__[a-zA-Z_][a-zA-Z_0-9]*__'}
    ,'method':    {'cn': 'hl_method',   'rs': '\\.\\w+'}
    ,'operator':  {'cn': 'hl_operator', 'words': ['%\\=','%','\\!\\=','\\=\\=','\\+\\=','-\\=','\\*\\=','&\\=','\\<\\=','\\>\\=','\\|\\=','\\*\\*','\\<\\<','\\>\\>','\\<\\>','\\=','\\+','\\-','\\*','\\/','\\<','\\>','&','\\|','\\^','~','\\:',',',';']}
    ,'keyword':   {'cn': 'hl_keyword',  'rs': '(?!\\W)(', 're': ')(?=\\W)', 'words': ['and','assert','break','class','continue','def','del','elif','else','except','exec','finally','for','from','global','if','import','in','is','lambda','not','or','pass','print','raise','return','try','while','with','yield','True','False']}
    ,'decorator': {'cn': 'hl_decorator','rs': '@[a-zA-Z_][a-zA-Z_0-9]*'}
    ,'word':      {'cn': 'hl_word',     'rs': '[a-zA-Z_][a-zA-Z_0-9]*'}
    ,'integer':   {'cn': 'hl_number',   'rs': ul.hl.re.num_int}
  }
 };

// highlight text (inserting 'span' tags with corresponding class names)
// hl_text - input text
// hl_dict - highlight format dictionary : {scheme1 : {element1: {params}, element2: {params} ...}, scheme2: {element1: {params}, element2: {params} ...} ...}
// hl_sch  - highlighting scheme name
function hl_text(hl_text, hl_dict, hl_sch) {
  var str_in = hl_text;  // incoming text
  var str_out = hl_text; // outcoming text
  if (hl_text && hl_sch && hl_dict && hl_dict[hl_sch]) {

    // if necessary, build new regular expression string and highlight classes array
    if (!hl_dict[hl_sch]['re_str']) {
      var re_str = '';     // regular expression string for current scheme
      var re_classes = []; // highlight classes names, corresponding to regular expression subexpression number
      // create regexp string
      for (var hl_cn in hl_dict[hl_sch]) { // iterate for all highlight classes
        var hl_class = hl_dict[hl_sch][hl_cn]; // current highlight class
        var re_str1 = '('; // starting grouping character '('
        if (hl_class.rs) re_str1 += hl_class.rs; // starting regexp
        if (hl_class.words) re_str1 += hl_class.words.join('|'); // if words list present
        if (hl_class.re) re_str1 += hl_class.re;     // ending regexp
        re_str1 += ')|';   // ending grouping character ')' and combining '|'
        // update highlight classes names array
        for (re_classes_cnt = re_str1.split(/[^\\]\([^\?]/gm).length; re_classes_cnt > 0; re_classes_cnt--)
          re_classes.push(hl_class.cn);
        re_str += re_str1; // update overall regular expression
      }
      re_str = re_str.slice(0, -1); // remove ending '|'
      hl_dict[hl_sch]['re_str'] = re_str;
      hl_dict[hl_sch]['re_classes'] = re_classes;
    }

    str_in = str_in.replace(/\n\r|\r\n|\r/gm, '\n'); // f'n IE and it's line breaks and regexps - replace \r with \n
    var re = new RegExp(hl_dict[hl_sch]['re_str'], 'gm'); // create regular expression object
    var index_prev = 0; // previous match ending index
    str_out = ''; // create 'highlighted' output string
    // regexp replacing loop
    do { // repeat for all matches
      re_res = re.exec(str_in);
      if (re_res) { // match found - replace it
        //search for match class name
        for (match_cnt = 1; (match_cnt < re_res.length) && (!re_res[match_cnt]); match_cnt++);
        if (index_prev < re_res.index) str_out += str_in.slice(index_prev, re_res.index);
        str_out += '<spanclass=\"' + hl_dict[hl_sch]['re_classes'][match_cnt - 1] + '\">' + // use 'spanclass' to hide space from &nbsp;
          str_in.slice(re_res.index, re_res.index + re_res[0].length) +
          '</span>';
        index_prev = re_res.index + re_res[0].length;
      }
    } while (re_res);
    // append non-processed piece of string
    if (index_prev <= str_in.length) str_out += str_in.slice(index_prev, -1);
    str_out = str_out.replace(/ /gm, '&nbsp;').replace(/[\n]/gm,'&nbsp;<br />').replace(/spanclass/gm, 'span class');
  }
  return(str_out);
}

// highlight content of HTML element(s) hl_elem (preferably PRE)
// original unhighlighted text will be stored to 'hl_text' attribute of every highlighted element
// if specific hl_sch (scheme name) given, it will be used as highlighting scheme,
//  else if hl_sch == 1, null, 'yes', it will be extracted from 'hl_code' attribute of every hl_elem,
//   else if hl_sch == 0, 'none', 'no', previously stored unhighlighted text will be shown
function hl_element(hl_elem, hl_dict, hl_sch, class_hl, class_unhl) {
  if (hl_elem && hl_dict && hl_elem.getAttribute && hl_elem.getAttribute(ul.hl.attr_code)) {
    if (!hl_sch) hl_sch = ul.params.hl;
    // if '1' - get it from 'hl_code' attribute of element
    if (hl_sch == 1) hl_sch = hl_elem.getAttribute(ul.hl.attr_code);
    // if '0' scheme given, unhighlight text
    if (hl_sch == 0) {
      // if original text stored, recover it to first TextNode of element
      if (hl_elem.getAttribute(ul.hl.attr_raw_text))
        hl_elem.innerHTML = hl_elem.getAttribute(ul.hl.attr_raw_text).replace(/ /gm, '&nbsp;').replace(/(\n\r)|(\r\n)|[\r\n]/gm,'&nbsp;<br />');
      // if given, set unhighlighted element class
      if (class_unhl) hl_elem.className = class_unhl;
    } else if (hl_dict[hl_sch]) {
        // if can, normalize element's TextNode subnodes to one subnode
        if (hl_elem.normalize) hl_elem.normalize();
        // remember original element text
        if (!hl_elem.getAttribute(ul.hl.attr_raw_text)) hl_elem.setAttribute(ul.hl.attr_raw_text, hl_elem.firstChild.nodeValue);
        // highlight element's content
        hl_elem.innerHTML = hl_text(hl_elem.getAttribute(ul.hl.attr_raw_text), hl_dict, hl_sch);
        // if given, set highlighted element class
        if (class_hl) hl_elem.className = class_hl;
      }
  }
}

// highlight source code present in childnodes of given parent node
function hl_childnodes(parent_node) {
  
  function hl_childnode(node_in) {
    if ((node_in.getAttribute) && (node_in.getAttribute(ul.hl.attr_code)) && (ul.hl.schemes[node_in.getAttribute(ul.hl.attr_code)]))
	  hl_element(node_in, ul.hl.schemes, ul.params.hl);
  }
  
  //dom_childnodes_walk(parent_node, hl_childnode);

  var el_cnt = 0;
  for (child_node in parent_node) {
    hl_element(parent_node[child_node], ul.hl.schemes, ul.params.hl);
    el_cnt++;
  }
  //document.getElementById('dbg').innerHTML = el_cnt;
}

// startup document translation and language select creation
function hl_onload() {

  function hl_sel_changed() {
    ul.params.hl = this.value;
    hl_childnodes(document.body.childNodes);
  }

  if ((hl_sel = document.getElementById(ul.hl.select_id)) != null) {
    var hl_opts = {'no':0, 'yes':1};
    for (var hl_opt in hl_opts)
      hl_sel.options.add(new Option(hl_opt, hl_opts[hl_opt], (hl_opts[hl_opt] == ul.params.hl), (hl_opts[hl_opt] == ul.params.hl)));
    hl_sel.onchange = hl_sel_changed;
  }
  hl_childnodes(document.body.childNodes);
}

ul.onload.push(hl_onload);
